/* compute the Adler-32 checksum of a data stream
 * Copyright (C) 1995-2006 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */
#include "adler32.h"

#define BASE 65521UL    /* largest prime smaller than 65536 */
#define NMAX 5552
/* NMAX is the largest n such that 255n(n+1)/2 + (n+1)(BASE-1) <= 2^32-1 */

#define MOD(a)      a %= BASE
#define MOD4(a)     a %= BASE


/* ========================================================================= */
unsigned long adler32(unsigned long adler, const unsigned char* buf,
                                           unsigned long len)
{
  unsigned long sum2;
  unsigned n;
  unsigned char i = 0;

  /* split Adler-32 into component sums */
  sum2 = (adler >> 16) & 0xffff;
  adler &= 0xffff;

  /* in case user likes doing a byte at a time, keep it fast */
  if (len == 1) {
    adler += buf[0];
    if (adler >= BASE)
      adler -= BASE;
    sum2 += adler;
    if (sum2 >= BASE)
      sum2 -= BASE;
    return adler | (sum2 << 16);
  }

  /* initial Adler-32 value (deferred check for len == 1 speed) */
  if (buf == ((unsigned char*)(0))) 
    return 1L;

  /* in case short lengths are provided, keep it somewhat fast */
  if (len < 16) {
    while (len--) {
      adler += *buf++;
      sum2 += adler;
    }
    if (adler >= BASE)
      adler -= BASE;
    MOD4(sum2);             /* only added so many BASE's */
    return adler | (sum2 << 16);
  }

  /* do length NMAX blocks -- requires just one modulo operation */
  while (len >= NMAX) {
    len -= NMAX;
    n = NMAX / 16;          /* NMAX is divisible by 16 */
    do {
      for ( i = 0; i < 16; i ++, buf++ ) {        /* 16 sums unrolled */
        adler += *buf;
        sum2 += adler;
      }
    } while (--n);
    MOD(adler);
    MOD(sum2);
  }

  /* do remaining bytes (less than NMAX, still just one modulo) */
  if (len) {                  /* avoid modulos if none remaining */
    while (len >= 16) {
      len -= 16;
      for ( i = 0; i < 16; i ++, buf++ ) {
        adler += *buf;
        sum2 += adler;
      }
    }
    while (len--) {
      adler += *buf++;
      sum2 += adler;
    }
    MOD(adler);
    MOD(sum2);
  }

  /* return recombined sums */
  return adler | (sum2 << 16);
}

